@@ -1,3 +1,7 @@
+"""This module implements the TftpPacketFactory class, which can take a binary
+buffer, and return the appropriate TftpPacket object to represent it, via the
+parse() method."""
+
 from TftpShared import *
 from TftpPacketTypes import *
 
