@@ -3,12 +3,12 @@ instance of the server, and then run the listen() method to listen for client
 requests. Logging is performed via a standard logging object set in
 TftpShared."""
 
-import socket, os, re, time, random
+import socket, os, time
 import select
 from TftpShared import *
 from TftpPacketTypes import *
-from TftpPacketFactory import *
-from TftpStates import *
+from TftpPacketFactory import TftpPacketFactory
+from TftpContexts import TftpContextServer
 
 class TftpServer(TftpSession):
     """This class implements a tftp server object. Run the listen() method to
@@ -94,7 +94,6 @@ class TftpServer(TftpSession):
 
                     log.debug("Read %d bytes" % len(buffer))
 
-                    recvpkt = tftp_factory.parse(buffer)
                     # Forge a session key based on the client's IP and port,
                     # which should safely work through NAT.
                     key = "%s:%s" % (raddress, rport)
@@ -107,7 +106,12 @@ class TftpServer(TftpSession):
                                                                timeout,
                                                                self.root,
                                                                self.dyn_file_func)
-                        self.sessions[key].start(buffer)
+                        try:
+                            self.sessions[key].start(buffer)
+                        except TftpException, err:
+                            deletion_list.append(key)
+                            log.error("Fatal exception thrown from "
+                                      "session %s: %s" % (key, str(err)))
                     else:
                         log.warn("received traffic on main socket for "
                                  "existing session??")
@@ -144,9 +148,17 @@ class TftpServer(TftpSession):
             for key in self.sessions:
                 try:
                     self.sessions[key].checkTimeout(now)
-                except TftpException, err:
+                except TftpTimeout, err:
                     log.error(str(err))
-                    deletion_list.append(key)
+                    self.sessions[key].retry_count += 1
+                    if self.sessions[key].retry_count >= TIMEOUT_RETRIES:
+                        log.debug("hit max retries on %s, giving up"
+                            % self.sessions[key])
+                        deletion_list.append(key)
+                    else:
+                        log.debug("resending on session %s"
+                            % self.sessions[key])
+                        self.sessions[key].state.resendLast()
 
             log.debug("Iterating deletion list.")
             for key in deletion_list:
